from structlinks.DataStructures.BinarySearchTree.BinarySearchTree import BinarySearchTree
from structlinks.DataStructures.DoublyLinkedList.DoublyLinkedList import DoublyLinkedList
from structlinks.DataStructures.Graph.Graph import Graph
from structlinks.DataStructures.LinkedList.LinkedList import LinkedList
from structlinks.DataStructures.Matrix.Matrix import Matrix
from structlinks.DataStructures.Queue.Queue import Queue
from structlinks.DataStructures.Heap.Heap import Heap, MaxHeap, MinHeap, PriorityFunction